#![allow(dead_code)]
use std::collections::HashMap;
use std::collections::hash_map::Entry;
use rand;
use hex;

/// Message type to distinguish
/// the different messages send via
/// dns
#[derive(Copy, Clone, Eq, PartialEq, Debug)]
pub enum MessageType {
    POLL,       // 1
    INIT,       // 2
    ACK,        // 3
    DATA,       // 4
    DATAFIN,    // 5
    UNKNOWN // _
}


impl MessageType {

    pub fn from_str(input: &str) -> Self {
        match input {
            "1" => MessageType::POLL,
            "2" => MessageType::INIT,
            "3" => MessageType::ACK,
            "4" => MessageType::DATA,
            "5" => MessageType::DATAFIN,
            _ => MessageType::UNKNOWN,
        }
    }

    /// maps the enum to a string (num)
    pub fn to_str(msg_type: MessageType) -> &'static str {
        match msg_type {
            MessageType::POLL       => "1",
            MessageType::INIT       => "2",
            MessageType::ACK        =>  "3",
            MessageType::DATA       => "4",
            MessageType::DATAFIN    => "5",
            MessageType::UNKNOWN    => "0"
        }
    }
}


/// Every DNS message is packed into
/// this Message Format
/// Implemented in DNS via:
/// [type].[id].[data].[domain].[tld]
#[derive(Debug)]
pub struct Message {
    pub msg_type: MessageType,
    pub msg_id: u16,
    pub data: Vec<u8>
}


impl<'a> Message {

    /// Craft a new message from data only, creating a unique `msg_id`
    pub fn new(msg_type: MessageType, data: Vec<u8>) -> Self {
        let msg_id = rand::random::<u16>();
        Message { msg_type, msg_id, data}
    }

    /// Specifiy the ID of the packet to generate several messages
    /// with the same ID
    pub fn with_id(msg_type: MessageType, msg_id: u16, data: Vec<u8>) -> Self {
        Message { msg_type, msg_id, data }
    }

    pub fn default() -> Self {
        Message {msg_type: MessageType::INIT, msg_id: rand::random::<u16>(), data: Vec::new() }
    }

    /// parse a message form a subdomain string
    pub fn from(subdomain: &str, full_domain: bool) -> Option<Self> {
        // Split the subdomain at every dot and parse into Message Struct
        let parts: Vec<&str> = subdomain.splitn(3, ".").collect();

        let msg_type = MessageType::from_str(parts[0]);
        let msg_id: u16 = parts[1].parse().unwrap();
        let mut data: Vec<u8> = Vec::new();
        let data_string: String;
        if full_domain {
            let tmp_data_string: Vec<&str> = parts[2].rsplitn(3, ".").collect();
            data_string = tmp_data_string[2].replace(".", "");
        } else {
            data_string = parts[2].replace(".", "");
        }
        if data_string != "00" {
            data = hex::decode(&data_string).unwrap();
        }
        Some(Message {msg_type, msg_id, data})
    }

    /// outputs the fields of the message struct in the
    /// subdomain order [type].[id].[data]
    pub fn to_string(&self) -> String {
            let mut subdomain = String::new();
            subdomain.push_str(MessageType::to_str(self.msg_type));
            subdomain.push_str(".");
            subdomain.push_str(&self.msg_id.to_string());
            subdomain.push_str(".");

            if self.data.len() == 0 {
                subdomain.push_str("00");
            } else {
                // split the encoded data into chunks of 63 chars
                // [type].[id].[data].[data]...[data]
                let encoded_data = hex::encode(&self.data);
                let chunked = self.split_chunks(&encoded_data);
                for chunk in chunked {
                    subdomain.push_str(chunk);
                    subdomain.push_str(".");
                }
                subdomain.pop();
            }
            subdomain
    }

    fn split_chunks(&self, data: &'a String) -> Vec<&'a str> {

        let chunks = data.as_bytes()
            .chunks(62)
            .map(std::str::from_utf8)
            .collect::<Result<Vec<&str>, _>>()
            .unwrap();
        return chunks;
    }
}


/// A handler the keeps a HashMap to track the unfinished
/// packets via the unique packet ID
#[derive(Debug)]
pub struct MessageQueue {
    pub packets: Vec<Vec<u8>>, // the packets that are already finished
    tmp_packets: HashMap<u16, Vec<u8>>,  // the packets that need more data
}


impl MessageQueue {
    pub fn new() -> Self {
        MessageQueue { packets: Vec::new(), tmp_packets: HashMap::new() }
    }


    /// Takes a `Message` as input and includes it into the hashmap
    /// according to the supplied ID
    pub fn insert_msg(&mut self, msg: Message) {

        match msg.msg_type {
            MessageType::DATA => {
                // For unfinished data packets, insert them
                // into the `tmp_packets` HashMap, as there is more
                // data to come
                match self.tmp_packets.entry(msg.msg_id) {
                    Entry::Occupied(mut entry) => { entry.get_mut().extend(&msg.data); }
                    Entry::Vacant(entry) => { entry.insert(msg.data); }
                };
            },
            MessageType::DATAFIN => {

                match self.tmp_packets.entry(msg.msg_id) {
                    Entry::Occupied(mut entry) => {
                        entry.get_mut().extend(&msg.data);
                        self.packets.push(entry.get().to_vec());
                        entry.remove();
                    }
                    Entry::Vacant(_) => {
                        if msg.data.len() > 0 {
                            self.packets.push(msg.data);
                        }
                    }
                };
            },
            _ => {}
        }
    }
}

