// package only crates
use crate::message::{Message, MessageType, MessageQueue};
use crate::dns::*;
use std::net::UdpSocket;
use std::str;
use std::thread;
use std::io::{Write, Read};
// using for the tun interface and parsing the packets
use tun;
// Channels for inter-thread communication
use std::sync::mpsc::{Sender, Receiver};
use std::sync::{mpsc, Arc, Mutex};
use std::collections::VecDeque;


/// Struct to keep track of the packets that need to
/// be returned from SERVER -> CLIENT
#[derive(Clone)]
pub struct ReturnPacket {
    pub data: Vec<u8>,
    pub pkt_id: u16
}

impl ReturnPacket {
    /// Construct a new packet with the supplied data and a unique `pkt_id`
    pub fn new() -> Self {
        ReturnPacket { data: Vec::new(), pkt_id: rand::random::<u16>() }
    }
}

#[derive(Clone)]
pub struct DnsTunnel {
    server: String,
    domain: String,
    port: u16,
    return_packet: ReturnPacket
}

impl<'a> DnsTunnel {

    /// Craft a new instance of a DNS Tunnel
    /// ip: IP of the DNS tunnel server
    /// port: port of the DNS tunnel server
    pub fn new(domain: String, server: String, port: u16) -> DnsTunnel {
        DnsTunnel { server, port, domain, return_packet: ReturnPacket::new() }
    }

    /// keeps open a listener to transfer data via DNS tunneling
    pub fn new_server(domain: String, port: u16) -> DnsTunnel {
        DnsTunnel { server: String::new(), port, domain, return_packet: ReturnPacket::new() }
    }

    pub fn connect(&self) -> Result<(), &str> {
        if !self.check_connection() {
            println!("[-] DNS Tunnel not reachable");
            return Err("");
        }
        println!("[+] Successfully connected to DNS Tunnel");
        Ok(())
    }

    pub fn serve(&mut self) {


        println!("[+] Started DNS tunnel, ready to receive data...");
        let socket = UdpSocket::bind(("0.0.0.0", self.port)).unwrap();


        // Starting TUN interface on the server side
        // Creating Tun interface with addr. 10.0.0.1
        let mut write_config = tun::Configuration::default();
        let mut read_config = tun::Configuration::default();
        write_config.address((10, 0, 0, 1))
               .netmask((255, 255, 255, 0))
               .mtu(1500)
               .layer(tun::Layer::L3)
               .up();
        read_config.address((10, 0, 0, 3))
               .netmask((255, 255, 255, 0))
               .mtu(1500)
               .layer(tun::Layer::L3)
               .up();


        println!("[*] Spawning tun interface");
        let mut read_dev = tun::create(&read_config).unwrap();   // tun1
        let mut write_dev = tun::create(&write_config).unwrap();  // tun0

        //let mut queue = Arc::new(RwLock::new(VecDeque::new()));
        let queue = Arc::new(Mutex::new(VecDeque::new()));
        let sender_queue = queue.clone();
        let recv_queue = queue.clone();

        // channels to communicate
        let (tx, rx): (Sender<Vec<u8>>, Receiver<Vec<u8>>) = mpsc::channel();


        // Create a new thread to handle the incoming packets (via DNS),
        // and write ehm to the tun interface
        thread::spawn(move ||

            //let mut buf = [0; 4096];
            loop {
                // receive the packet from the receiver (coming from DNS)
                let packet_from_dns = rx.recv().unwrap();
                // write that packet into the tun interface
                // (routing done via iptables)
                write_dev.write(&packet_from_dns).expect("[-] Failed to write packet to device");
            }
        );


        // Create a second thread to handle the incoming packets (from
        // the tun interface) and send em back to the DNS client
        // (via the polling interface)
        thread::spawn(move || {
            // create a new thread for the tunnel
            let mut buf = [0; 4096];
            loop {
                let amount: usize;
                amount = read_dev.read(&mut buf).unwrap();
                if amount ==0 {
                    continue;
                }

                let packet = &buf[0 .. amount];
                /*
                match ip::Packet::new(packet) {
                    Ok(ip::Packet::V4(pkt)) => {
                        println!("[SERVER -> CLIENT]: Captured packet: Src: {:?} Dst: {:?}", pkt.source(), pkt.destination());
                    }
                    _ => {
                    }
                }
                */
                sender_queue.lock().unwrap().push_back(packet.to_vec());
            }
        });


        // starting main loop
        let mut poll_status = PollStatus::NONE;

        let mut message = String::new();

        // A message queue to store the tmp_messages
        let mut msg_queue = MessageQueue::new();

        // The main DNS Tunnel loop that works with the different packet types
        loop {
            let mut response_buffer = BytePacketBuffer::new();

            // receive into response buffer
            let (_, src) = socket.recv_from(&mut response_buffer.buf).unwrap();

            // decode the response as DNS packet
            let packet = DnsPacket::from_bytes(&mut response_buffer);

            // if there is no at least one question or not of type TXT
            if packet.questions.len() < 1 ||
                packet.questions[0].qtype != QueryType::TXT {
                    continue;
            }

            // Exfiltrate the Data
            let data = &packet.questions[0].name;
            if let Some(recv_msg) = Message::from(data, true) {
                // Return DNS Packet
                let mut return_packet = DnsPacket::empty();

                match recv_msg.msg_type {
                    MessageType::INIT => {
                        return_packet = DnsTunnel::build_response(&packet, MessageType::ACK, None, None);
                    },
                    MessageType::DATA => {
                        // For data transfer from CLIENT -> SERVER
                        msg_queue.insert_msg(recv_msg);  // insert into the message queue
                        return_packet = DnsTunnel::build_response(&packet, MessageType::ACK, None, None);
                    },
                    MessageType::DATAFIN => {
                        // Data transfer CLIENT -> SERVER finished
                        msg_queue.insert_msg(recv_msg); // insert the last packet

                        // take all the packets and send ehm to the
                        // tun interface
                        for packet in &msg_queue.packets {
                            tx.send(packet.to_vec()).expect("[-] Failed to send packet via channel");
                        }
                        msg_queue.packets.clear();
                    },
                    MessageType::POLL => {
                        // For data transfer from SERVER -> CLIENT
                        match poll_status {
                            PollStatus::NONE => {
                                // Enter Polling Mode
                                return_packet = DnsTunnel::build_response(&packet, MessageType::POLL, None, None);

                                // create new return packet
                                self.return_packet = ReturnPacket::new();


                                // if the queue is empty, polling is done
                                let queue = recv_queue.lock().unwrap(); // make mutex available
                                if queue.is_empty() {
                                    poll_status = PollStatus::NONE;
                                } else {
                                    poll_status = PollStatus::STARTED;
                                }
                            },
                            PollStatus::STARTED => {

                                // create placeholder variables
                                let chunk: Vec<u8>;
                                let pkt_id: u16;
                                // max length for the data in one DNS packet
                                let max_len: usize = (255 - self.domain.len() - 3 - 5 - 3) / 2;
                                // unllock the global DeQueue
                                let mut queue = recv_queue.lock().unwrap(); // make mutex available
                                // if the global buffer is already empty,
                                // get a new item from the queue and
                                // send this in chunks to the client
                                if self.return_packet.data.len() == 0 && queue.len() >= 1 {
                                    self.return_packet.data = queue.pop_front().unwrap();
                                    //self.return_packet.data = queue_rx.recv().unwrap();
                                }
                                if self.return_packet.data.len() >= max_len {
                                    chunk = self.return_packet.data[..max_len].to_vec();
                                    pkt_id = self.return_packet.pkt_id;
                                    self.return_packet.data = self.return_packet.data[max_len..].to_vec();
                                    return_packet = DnsTunnel::build_response(&packet, MessageType::DATA, Some(pkt_id), Some(chunk));
                                } else {
                                    // send the rest via the DATAFIN packet (<max_len)
                                    chunk = self.return_packet.data.to_vec();
                                    pkt_id = self.return_packet.pkt_id;
                                    return_packet = DnsTunnel::build_response(&packet, MessageType::DATAFIN, Some(pkt_id), Some(chunk));
                                    poll_status = PollStatus::NONE;
                                    // reset the ReturnPacket
                                    self.return_packet = ReturnPacket::new();
                                }


                            }
                        }
                    },
                    _ => {
                        message.push_str(&data);
                    }

                }
                // return the response Packet
                let res_buffer = DnsTunnel::get_buffer(&mut return_packet);
                socket.send_to(res_buffer.get_range(0, res_buffer.pos()).unwrap(), src).unwrap();
            }
        }
    }


    pub fn poll(&self) -> Option<Message> {
        // create a new POLL message
        let poll_msg = Message::new(MessageType::POLL, Vec::new());

        // send the msg and receive the outout
        if let Some(msg) = self.send_and_recv_message(poll_msg) {
            return Some(msg);
        }
        return None;
    }
    pub fn send_bytes(&self, bytes: Vec<u8>) {
        // The MAX_LEN = 255 of the full domain
        const MAX_LEN: usize = 255;
        // substract len of domain, 3 (dots), 6 (id & type)
        // devided by two as the encoded representation of a byte takes to
        // characters
        let data_len: usize = (MAX_LEN - self.domain.len() - 3 - 6) / 2;

        // if the size of the data execeeds
        if bytes.len() > data_len {
            // split the data in chunks of 63 bytes
            let chunks = bytes
                .chunks(data_len)
                .collect::<Vec<&[u8]>>();

            let queue_id = rand::random::<u16>();

            // send the chunks
            for chunk in chunks {
                // craft and send the chunk
                let tmp_msg = Message::with_id(MessageType::DATA, queue_id, chunk.to_vec());
                self.send_message(tmp_msg);
            }
            // send a DATAFIN packet if every chunk is send
            let tmp_msg = Message::with_id(MessageType::DATAFIN, queue_id, Vec::new());
            self.send_message(tmp_msg);
        } else {
            let single_msg = Message::new(MessageType::DATAFIN, bytes);
            self.send_message(single_msg);
        }

    }


    /// Check if the remote server answers a simple `TUN` message with the
    /// answer `ACK`, this means that the server is available and can
    /// send back data.
    fn send_and_recv_message(&self, msg: Message) -> Option<Message> {

        // Craft new request from `msg`
        let mut pkt = self.build_request(msg);

        // send the request to the DNS Tunnel Server
        let resp = pkt.send((&self.server, self.port));

        // Check for any answers to that query
        if resp.answers.len() < 1 {
            return None;
        }
        // Parse the answer
        match &resp.answers[0] {
            DnsRecord::TXT {
                ref domain,
                len: _,
                ref txt,
                ttl: _
            } => {
                return Message::from(txt, false);
            }
            _ => {
                return None;
            }
        };
    }
    /// Check if the remote server answers a simple `TUN` message with the
    /// answer `ACK`, this means that the server is available and can
    /// send back data.
    fn send_message(&self, msg: Message) -> bool {

        // Craft new request from `msg`
        let mut pkt = self.build_request(msg);

        // send the request to the DNS Tunnel Server
        let resp = pkt.send((&self.server, self.port));

        // Check for any answers to that query
        if resp.answers.len() < 1 {
            return false;
        }
        // Parse the answer
        match &resp.answers[0] {
            DnsRecord::TXT {
                ref domain,
                len: _,
                ref txt,
                ttl: _
            } => {
                if let Some(return_message) = Message::from(txt, false) {
                    if return_message.msg_type == MessageType::ACK {
                        return true;
                    }
                }
            }
            _ => {
                    return false;
            }
        };
        return false;
    }

    /// Sends a `INIT` message to the server to check if the servern
    /// returns a `ACK`
    fn check_connection(&self) -> bool {
        // create a INIT message with empty data field
        let init_msg = Message::new(MessageType::INIT, Vec::new());
        self.send_message(init_msg)
    }

    /// Creates a DNS reponse packet, taking a `DnsPacket` as reference
    /// and a message struct that is send in ther TXT response field
    fn build_response(ref_pkt: &DnsPacket,
            msg_type: MessageType,
            msg_id: Option<u16>,
            data: Option<Vec<u8>>) -> DnsPacket {
        // Create a empty DnsPacket as base, and fill it with
        // the needed data from the reference packet
        let mut packet = DnsPacket::empty();
        packet.header.id = ref_pkt.header.id;
        packet.header.recursion_desired = ref_pkt.header.recursion_desired;
        packet.header.recursion_available = true;
        packet.header.query_response = true;
        packet.header.result_code = ResultCode::NOERROR;

        let domain = &ref_pkt.questions[0].name;

        let mut msg = Message::default();

        if let Some(msg_id) = msg_id {
            msg.msg_type = msg_type;
            msg.msg_id = msg_id;
        } else {
            msg.msg_type = msg_type;
        }

        if let Some(data) = data {
            msg.data = data;
        }
        let return_data = msg.to_string();
        let answer = DnsRecord::TXT {
            domain: domain.to_string(),
            len: return_data.len() as u8,
            ttl: 7766,
            txt: return_data
        };
        packet.answers.push(answer);
        return packet;

    }
    fn build_request(&self, msg: Message) -> DnsPacket {
        let mut packet = DnsPacket::empty();
        let full_domain = format!("{}.{}", &msg.to_string(), self.domain);
        // build a DNS TXT query out of the message
        let question = DnsQuestion{ name: full_domain, qtype: QueryType::TXT};
        packet.questions.push(question);

        packet
    }

    fn get_buffer(pkt: &mut DnsPacket) -> BytePacketBuffer {
        let mut res_buffer = BytePacketBuffer::new();
        pkt.write(&mut res_buffer).unwrap();
        return res_buffer;
    }
}


enum PollStatus {
    NONE,
    STARTED,
}
